import json
import numpy as np
import dicom2nifti
import dicom2nifti.settings as settings

from datetime import datetime
from dicom2nifti.convert_dir import _remove_accents
from loguru import logger
from operator import itemgetter
from pydicom import dcmread
from pathlib import Path
from shutil import copyfile, SameFileError

from mircato_stats.configs.logging import timer


# Set up the constant values that we will need for the dicom header data
ORIENTATIONS = {
    "COR": [1, 0, 0, 0, 0, -1],
    "SAG": [0, 1, 0, 0, 0, -1],
    "AX": [1, 0, 0, 0, 1, 0],
}
MIN_FILE_COUNT = 40
# Dictionary of tag names and tag keys, and if the tag is a sequence
IMPORTANT_ATTRIBUTES = {
    "PatientID": "mrn",
    "AccessionNumber": "accession",
    "SeriesDescription": "series_name",
    "SeriesNumber": "series_number",
    "ImageOrientationPatient": "ct_direction",
    "ImageType": "image_type",
    "PatientSex": "sex",
    "PatientSize": "height_at_scan_m",
    "PatientWeight": "weight_at_scan_kg",
    "PregnancyStatus": "pregnancy_status",
    "PatientAge": "age_at_scan",
    "PatientBirthDate": "birthday",
    "AcquisitionDate": "scan_date",
    "PixelSpacing": None,
    "SliceThickness": "slice_thickness_mm",
    "Rows": "num_rows",
    "Columns": "num_columns",
    "Manufacturer": "manufacturer",
    "ManufacturerModelName": "manufacturer_model",
    "KVP": "kvp",
    "SequenceName": "sequence_name",
    "ProtocolName": "protocol_name",
    "ContrastBolusAgent": "contrast_bolus_agent",
    "ContrastBolusRoute": "contrast_bolus_route",
    "MultienergyCTAcquisitionSequence": "multienergy_ct",
    "ProcedureCodeSequence": None,
    "StudyInstanceUID": "study_uid",
    "SeriesInstanceUID": "series_uid",
}


# Set the dicom to nifti settings
settings.disable_validate_slice_increment()
settings.enable_resampling = True
settings.resampling_order = 1
settings.reorient_nifti = True


class NotADicomFolderError(Exception):
    "Raised when the folder is not a valid dicom folder"

    pass


class InsufficientDicomFilesError(Exception):
    "Raised when there are not enough dicom files in a folder"

    pass


class IncompleteDicomReferenceError(Exception):
    "Raised when the dicom information is incomplete"

    pass


class DicomFolder(Path):
    "Class representing a Dicom Folder to be converted and store a reference"

    def _check_folder(self) -> None:
        "Check the contents of the folder to ensure it is a valid dicom folder"
        min_file_count = MIN_FILE_COUNT
        if not self.is_dir():
            raise NotADicomFolderError
        dicom_files = list(self.glob("*.dcm"))
        if not dicom_files:
            dicom_files = list(self.glob("*"))
        if len(dicom_files) < min_file_count:
            raise InsufficientDicomFilesError
        self.reference_dicom = dicom_files[0]
        self.reference_dicom_name = dicom_files[0].parts[-1]
        self.series_numbers = set(
            [getattr(dcmread(dicom), "SeriesNumber") for dicom in dicom_files]
        )
        self.series_name = _remove_accents(
            getattr(dcmread(self.reference_dicom), "SeriesDescription", "")
        )

    def _generate_reference_dict(self) -> None:
        "Load reference dicom header and generate a dictionary containing critical data"
        ref_header = dcmread(self.reference_dicom)
        naming_keys = ["mrn", "accession", "series_name"]
        ref_data = {}
        important_attributes = IMPORTANT_ATTRIBUTES
        orientations = ORIENTATIONS
        for tag, column in important_attributes.items():
            value = getattr(ref_header, tag, None)
            if value == "":
                value = None
            if value is not None:
                match tag:
                    case "SeriesDescription":
                        value = self.series_name
                    case "PatientID":
                        value = str(value).zfill(8)
                    case "ImageOrientationPatient":
                        value = None
                        orientation = np.round(getattr(ref_header, tag))
                        for k, direction in orientations.items():
                            if np.array_equal(orientation, direction):
                                value = k
                                break
                    case "ImageType":
                        value = "_".join(value)
                    case tag if tag in ["PatientBirthDate", "AcquisitionDate"]:
                        value = datetime.strptime(value, "%Y%m%d").date()
                        value = value.strftime("%Y-%m-%d")
                    case "PatientAge":
                        value = int(value[:-1])
                    case "PixelSpacing":
                        length, width = value
                        ref_data["length_mm"] = length
                        ref_data["width_mm"] = width
                        continue
                    case "PatientSex":
                        if value == "M":
                            value = 0
                        elif value == "F":
                            value = 1
                        else:
                            value = None
                    case tag if tag in ["ContrastBolusAgent", "ContrastBolusRoute"]:
                        value = 1
                    case "MultienergyCTAcquisitionSequence":
                        value = 1
                    case "PregnancyStatus":
                        value = int(value)
                        if value == 4:
                            value = None
                        elif value == 1:
                            value = 0
                        elif value == 2:
                            value = 1
                        elif value == 3:
                            value = 2
                    case tag if tag in ["SeriesNumber", "Rows", "Columns"]:
                        value = int(value)
                    case tag if tag in ["PatientSize", "PatientWeight"]:
                        value = float(value)
                    case tag if tag in ["SequenceName", "ProtocolName"]:
                        value = str(value).lower()
                    case "ProcedureCodeSequence":
                        seq = value[0]
                        ref_data["procedure_code"] = seq.CodeValue
                        ref_data["procedure_desc"] = seq.CodeMeaning
                        continue
                    case _:
                        value = value
                ref_data[column] = value
            elif tag == "ProcedureCodeSequence":
                ref_data["procedure_code"] = None
                ref_data["procedure_desc"] = None
            elif tag == "PixelSpacing":
                ref_data["length_mm"] = None
                ref_data["width_mm"] = None
            else:
                ref_data[column] = None
        # Perform two quick checks for data existence
        ref_data["is_mip"] = ("mip" in ref_data.get("series_name", "").lower()) or (
            ref_data.get("slice_thickness_mm", 5) > 5
        )
        self.ref_data = ref_data
        if all([x is None for x in ref_data.values()]) or any(
            [ref_data.get(key) is None for key in naming_keys]
        ):
            raise IncompleteDicomReferenceError

    @timer
    def _convert(self, nifti_dir: Path) -> None:
        """Convert the dicoms to nifti after checks have been passed
        Parameters
        ----------
        nifti_dir : Path
            The directory where the nifti file will be saved
        """
        if not nifti_dir.exists():
            nifti_dir.mkdir(parents=True)

        try:
            copyfile(self.reference_dicom, nifti_dir / self.reference_dicom_name)
        except SameFileError:  # This sometimes gets raised, not sure why
            pass
        # Write the reference dict to a json file called header_info.json
        with (nifti_dir / "header_info.json").open("w") as f:
            json.dump(self.ref_data, f, indent=4)
        # Convert the dicoms to nifti
        dicom2nifti.convert_directory(self, nifti_dir, compression=True, reorient=True)

    def convert_to_nifti(self, output_dir: Path, only_ax: bool, no_mip: bool) -> None:
        """Converts the dicom folder to a nifti file.
        Parameters
        ----------
        output_dir : Path
            The base directory where the nifti file will be saved
        only_ax : bool
            If True, only will be saved if the series is an axial view
        no_mip : bool
            If True, MIP series will not be saved
        """
        extra_dict = {
            "key": "dicom_to_nifti",
            "input_dir": str(self),
            "converted": False,
            "output_dir": None,
            "output_files": None,
            "failed_reason": None,
        }
        try:
            self._check_folder()
            self._generate_reference_dict()
            if only_ax and self.ref_data["ct_direction"] != "AX":
                extra_dict["failed_reason"] = "not_axial"
                logger.error(
                    f"{self} is not an axial scan and only_ax = True, will not convert",
                    extra=extra_dict,
                )
                return
            if no_mip and self.ref_data["is_mip"]:
                extra_dict["failed_reason"] = "likely_mip"
                logger.error(
                    f"{self} is likely a MIP scan and no_mip = True, will not convert",
                    extra=extra_dict,
                )
                return
            mrn, accession, series_name = itemgetter("mrn", "accession", "series_name")(
                self.ref_data
            )
            first_two_mrn_digits = mrn[:2]
            nifti_dir = (
                output_dir / first_two_mrn_digits / mrn / accession / series_name
            )
            _, convert_time = self._convert(nifti_dir)
            nifti_files = [
                f"{n}_{self.series_name}.nii.gz" for n in self.series_numbers
            ]
            # Update the extra_dict with the output directory and files
            extra_dict["converted"] = True
            extra_dict["output_dir"] = str(nifti_dir)
            extra_dict["output_files"] = nifti_files
            logger.success(
                f"{self} converted to nifti in {convert_time:.2f} seconds. Saved to {nifti_dir}",
                extra=extra_dict,
            )
        except NotADicomFolderError:
            extra_dict["failed_reason"] = "not_a_directory"
            logger.error(f"{self} is not a directory", extra=extra_dict)
        except InsufficientDicomFilesError:
            extra_dict["failed_reason"] = "too_few_files"
            logger.error(
                f"Less than {MIN_FILE_COUNT} DICOM files found in {self}",
                extra=extra_dict,
            )
        except IncompleteDicomReferenceError:
            extra_dict["failed_reason"] = "incomplete_reference"
            logger.error(
                f"Dicom reference data is incomplete for {self}", extra=extra_dict
            )

        except SameFileError as e:
            extra_dict["failed_reason"] = "same_file_error"
            logger.warning(
                f"Nifti file already exists for {self}, will not convert. Error: {str(e)}",
                extra=extra_dict,
            )
        except dicom2nifti.exceptions.ConversionValidationError:
            extra_dict["failed_reason"] = "conversion_validation_error"
            logger.error(
                f"Dicom conversion could not be validated for {self}", extra=extra_dict
            )
        except PermissionError as e:
            extra_dict["failed_reason"] = "permission_error"
            logger.error(
                f"Dicom folder {self} could not be accessed due to a permission error {str(e)}",
                extra=extra_dict,
            )
        except Exception as e:
            extra_dict["failed_reason"] = type(e).__name__
            logger.error(
                f"A {type(e).__name__} error occured while converted the dicom folder {str(e)}",
                extra=extra_dict,
            )



torch_model_configs = {
    "total": {
        "name": "total",
        "spacing": [1.5, 1.5, 1.5],
        "patch_size": [128, 128, 128],
        "percentile_005": -1017,
        "percentile_995": 882,
        "mean": -260.3240051269531,
        "std": 477.13525390625,
        "output_map": {
            "spleen": 1,
            "kidney_right": 2,
            "kidney_left": 3,
            "gallbladder": 4,
            "liver": 5,
            "stomach": 6,
            "pancreas": 7,
            "adrenal_gland_right": 8,
            "adrenal_gland_left": 9,
            "lung_upper_lobe_left": 10,
            "lung_lower_lobe_left": 11,
            "lung_upper_lobe_right": 12,
            "lung_middle_lobe_right": 13,
            "lung_lower_lobe_right": 14,
            "esophagus": 15,
            "trachea": 16,
            "thyroid_gland": 17,
            "small_bowel": 18,
            "duodenum": 19,
            "colon": 20,
            "bladder": 21,
            "prostate": 22,
            "sacrum": 23,
            "vertebrae_S1": 24,
            "vertebrae_L5": 25,
            "vertebrae_L4": 26,
            "vertebrae_L3": 27,
            "vertebrae_L2": 28,
            "vertebrae_L1": 29,
            "vertebrae_T12": 30,
            "vertebrae_T11": 31,
            "vertebrae_T10": 32,
            "vertebrae_T9": 33,
            "vertebrae_T8": 34,
            "vertebrae_T7": 35,
            "vertebrae_T6": 36,
            "vertebrae_T5": 37,
            "vertebrae_T4": 38,
            "vertebrae_T3": 39,
            "vertebrae_T2": 40,
            "vertebrae_T1": 41,
            "vertebrae_C7": 42,
            "vertebrae_C6": 43,
            "vertebrae_C5": 44,
            "vertebrae_C4": 45,
            "vertebrae_C3": 46,
            "vertebrae_C2": 47,
            "vertebrae_C1": 48,
            "heart": 49,
            "aorta": 50,
            "pulmonary_vein": 51,
            "brachiocephalic_trunk": 52,
            "subclavian_artery_right": 53,
            "subclavian_artery_left": 54,
            "carotid_artery_right": 55,
            "carotid_artery_left": 56,
            "brachiocephalic_vein_left": 57,
            "brachiocephalic_vein_right": 58,
            "atrial_appendage_left": 59,
            "superior_vena_cava": 60,
            "inferior_vena_cava": 61,
            "portal_and_splenic_vein": 62,
            "iliac_artery_left": 63,
            "iliac_artery_right": 64,
            "iliac_vena_left": 65,
            "iliac_vena_right": 66,
        },
    },
    "body": {
        "name": "body",
        "spacing": [1.5, 1.5, 1.5],
        "patch_size": [128, 128, 128],
        "percentile_005": -985,
        "percentile_995": 1411,
        "mean": -43.4468879699707,
        "std": 355.778564453125,
        "output_map": {"body": 1},
    },
    "tissues": {
        "name": "tissues",
        "spacing": [1.5, 1.5, 1.5],
        "patch_size": [128, 128, 128],
        "percentile_005": -206,
        "percentile_995": 154,
        "mean": -33.02885055541992,
        "std": 74.95191192626953,
        "output_map": {"subq_fat": 1, "visc_fat": 2, "skel_muscle": 3},
    },
}
# Calculate the crop threshold using model vals
for config in torch_model_configs.values():
    config["crop_threshold"] = (config["percentile_005"] - config["mean"]) / config[
        "std"
    ]

nnunet_model_configs = {
    # Fill in configs here later
}



import numpy as np

from mircato_stats.configs.models import torch_model_configs
from mircato_stats.configs.logging import timer
from mircato_stats.statistics.nifti import NiftiMircato
from mircato_stats.statistics.utils import _calc_intensity_stats, _calc_shape_stats


@timer
def calculate_total_segmentation_stats(
    nifti: NiftiMircato,
) -> tuple[dict[str:float], dict[str:int]]:
    """Calculate the statistics for the total segmentation
    Parameters
    ----------
    nifti : NiftiMircato
        The nifti file to calculate statistics for
    Returns
    -------
    dict[str: float]
        The statistics for the total segmentation
    dict[str: int]
        The the vertebral midlines calculated from the total segmentation
    """
    # Set up the two output maps
    total_map = torch_model_configs["total"]["output_map"]
    vert_map = {k: v for k, v in total_map.items() if "vertebrae" in k}
    # Calculate the stats for the total segmentation
    shape_stats = _calc_shape_stats(nifti.total_seg)
    intensity_stats = _calc_intensity_stats(nifti.original_ct, nifti.total_seg)
    seg_labels = shape_stats.GetLabels()
    total_stats = {}
    # Loop through the shape and intensity stats for each structure
    for name, label in total_map.items():
        if label in seg_labels:
            volume = round(shape_stats.GetPhysicalSize(label) / 1000, 1)
            intensity = round(intensity_stats.GetMean(label), 2)
            total_stats[f"{name}_volume_cm3"] = volume
            total_stats[f"{name}_intensity"] = intensity
    # Calculate the vertebral midlines
    vert_midlines = {}
    for name, label in vert_map.items():
        if label in seg_labels:
            indices = shape_stats.GetIndexes(label)
            z_indices = indices[2::3]
            vert_midlines[f"{name}_midline"] = int(np.median(z_indices))

    return total_stats, vert_midlines



import os
import SimpleITK as sitk
from threadpoolctl import threadpool_limits


def set_num_threads(num_threads: int) -> None:
    """Set the number of threads available for every linear algebra related process (Numpy, Pytorch, SimpleITK, etc)
    :param num_threads: the number of threads
    :return: None
    """
    thread_vars = [
        *[f"{x}_NUM_THREADS" for x in ["OMP", "OPENBLAS", "MKL", "NUMEXPR"]],
        "VECLIB_MAXIMUM_THREADS",
    ]
    thread_dict = {var: str(num_threads) for var in thread_vars}
    os.environ.update(thread_dict)
    sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(num_threads)
    limiter = threadpool_limits(num_threads)



import SimpleITK as sitk
import numpy as np


def _calc_shape_stats(seg: sitk.Image) -> sitk.LabelShapeStatisticsImageFilter:
    """Calculate the shape stats for a segmentation
    Parameters
    ----------
    seg : sitk.Image
        The segmentation to calculate shape stats for
    Returns
    -------
    sitk.LabelShapeStatisticsImageFilter
        The executed shape stats
    """
    shape_stats = sitk.LabelShapeStatisticsImageFilter()
    shape_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    shape_stats.ComputeOrientedBoundingBoxOn()
    shape_stats.ComputePerimeterOn()
    shape_stats.Execute(seg)
    return shape_stats


def _calc_intensity_stats(
    image: sitk.Image, seg: sitk.Image
) -> sitk.LabelIntensityStatisticsImageFilter:
    """Calculate intensity stats for a segmentation using the reference image
    Parameters
    ----------
    image : sitk.Image
        The reference image
    seg : sitk.Image
        The segmentation of the reference image
    Returns
    -------
    sitk.LabelIntensityStatisticsImageFilter
        The executed intensity stats
    """
    intensity_stats = sitk.LabelIntensityStatisticsImageFilter()
    intensity_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    intensity_stats.Execute(seg, image)
    return intensity_stats


def _calculate_3d_volumes_and_intensities(
    image: sitk.Image,
    seg: sitk.Image,
    output_map: dict,
    prefix: str = "",
    endpoints: tuple = (0, None),
) -> dict:
    """Calculate the total volume and average intensity for structures in a region of 3d image
    :param image: the original image
    :param seg: the segmentation of the image
    :param output_map: the dictionary containing the output segmentation label map
    :param prefix: prefix to the statistics name, default is empty ''. Useful if measuring a specific region
    :param endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
        which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
    :return: a dictionary containing the volume in cm3 and average intensity for each segmentation in the output map
        found in the image. If the segmentation does not exist in the image, the return value will be None
    """
    volumes = _calculate_3d_volumes(seg, output_map, prefix, endpoints)
    intensities = _calculate_3d_intensities(image, seg, output_map, prefix, endpoints)
    if len(volumes) == len(intensities):
        stats = {}
        for (vol_key, vol_val), (inten_key, inten_val) in zip(
            volumes.items(), intensities.items()
        ):
            ord_stats = {vol_key: vol_val, inten_key: inten_val}
            stats.update(ord_stats)
    else:
        stats = {**volumes, **intensities}
    return stats


def _calculate_3d_volumes(
    seg: sitk.Image, output_map: dict, prefix: str = "", endpoints: tuple = (0, None)
) -> dict:
    """Calculate the total volume for each segmentation in a sitk image
    :param seg: the sitk image containing the segmentation
    :param output_map: the dictionary containing the output segmentation label map
    :param prefix: prefix to the statistics name, default is empty ''. Useful if measuring a specific region
    :param endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
        which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
    :return: a dictionary containing the volumes for each segmentation
    """
    seg = _slice_images(endpoints, seg)  # type: ignore
    shape_stats = _calc_shape_stats(seg)
    seg_labels = (
        shape_stats.GetLabels()
    )  # This is all labels that were found in the image
    volumes = {}
    for name, label in output_map.items():
        volume = None
        if label in seg_labels:
            volume = round(
                shape_stats.GetPhysicalSize(label) / 1000, 1
            )  # This gives the volume in cm3 for CT
        volumes[f"{prefix}{name}_volume_cm3"] = volume
    return volumes


def _calculate_3d_intensities(
    image: sitk.Image,
    seg: sitk.Image,
    output_map: dict,
    prefix: str = "",
    endpoints: tuple = (0, None),
) -> dict:
    """Calculate the average intensity for each segmentation in the reference image
    :param image: the reference image
    :param seg: the segmentation image
    :param output_map: the dictionary containing the output segmentation label map
    :param prefix: prefix to the statistics name, default is empty ''. Useful if measuring a specific region
    :param endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
        which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
    :return: a dictionary containing the average intensities for each segmentation
    """
    image, seg = _slice_images(endpoints, [image, seg])
    intensity_stats = _calc_intensity_stats(image, seg)
    seg_labels = intensity_stats.GetLabels()
    intensities = {}
    for name, label in output_map.items():
        intensity = None
        if label in seg_labels:
            intensity = round(intensity_stats.GetMean(label), 2)
        intensities[f"{prefix}{name}_average_intensity"] = intensity
    return intensities


def _calculate_2d_areas(
    seg: sitk.Image,
    output_map: dict,
    slice_idx: int,
    prefix: str,
    get_perimeter: bool = False,
) -> dict:
    """Calculate the area of each label in the segmentation within a 2d slice
    :param seg: the segmentation
    :param output_map: the label map
    :param slice_idx: the slice of the segmentation to measure
    :param prefix: the label of the slice (L1, L3, etc)
    :param get_perimeter: Measure the perimeters from the shape stats
    :return: a dictionary containing the area of each label found within the segmentation
    """
    seg_slice = seg[:, :, slice_idx]
    shape_stats = _calc_shape_stats(seg_slice)
    seg_labels = shape_stats.GetLabels()
    areas = {}
    for name, label in output_map.items():
        area = None
        if label in seg_labels:
            area = round(
                shape_stats.GetPhysicalSize(label) / 100, 1
            )  # This will be area in cm2 for CT
        areas[f"{prefix}{name}_area_cm2"] = area
        if get_perimeter and (label in seg_labels):
            # Note that the shape touches the border if at least 5 percent of the perimeter is on the border
            border_ratio = shape_stats.GetPerimeterOnBorderRatio(label) * 100
            raw_perim = (
                shape_stats.GetPerimeter(label) / 10
            )  # This will be perimeter in cm
            ellipse_perim = _calc_ellipsoid_perimeter(
                shape_stats.GetEquivalentEllipsoidDiameter(label)
            )
            circ_perim = shape_stats.GetEquivalentSphericalPerimeter(label) / 10
            areas[f"{prefix}{name}_border_ratio"] = round(border_ratio, 1)
            areas[f"{prefix}{name}_total_perimeter_cm"] = round(raw_perim, 1)
            areas[f"{prefix}{name}_ellipse_perimeter_cm"] = round(ellipse_perim, 1)
            areas[f"{prefix}{name}_circle_perimeter_cm"] = round(circ_perim, 1)
    return areas


def _slice_images(endpoints: tuple, images: sitk.Image | list) -> sitk.Image | list:
    """Slice any number of matching images to the same region using endpoints
    :param endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
        which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
    :param images: a single sitk.Image or list of sitk.Images -> must be the same shape or error is raised
    """
    start, end = endpoints
    is_single_image = isinstance(images, sitk.Image)
    if end is None:
        if is_single_image:
            end = images.GetSize()[
                -1
            ]  # This is the full length of the image along the z-axis
        else:
            # Check that all images are the same size
            image_sizes = [img.GetSize() for img in images]
            assert image_sizes.count(image_sizes[0]) == len(image_sizes), ValueError(
                "All images in list must be the same shape"
            )
            end = images[0].GetSize()[-1]  # All images are assumed to be the same size
    if is_single_image:
        images = images[:, :, start:end]
        return images
    else:
        sliced_images = [image[:, :, start:end] for image in images]
        return sliced_images


def _calc_ellipsoid_perimeter(diameters: tuple) -> float:
    """Calculate the estimated ellipsoid diameters from a shape_stats.GetEquivalentEllipsoidDiameter result.
    Uses the Ramanujan estimation.
    :param diameters: the result from shape_stats
    :return the perimeter in cm
    """
    minor, major = diameters
    major /= 2
    minor /= 2
    perim = np.pi * (
        3 * (major + minor) - np.sqrt((3 * major + minor) * (major + 3 * minor))
    )
    return perim / 10  # This will be in cm


def _filter_largest_components(image, labels_of_interest):
    """Filter the largest connected component for each label in the image
    :param image: the input image
    :param labels_of_interest: the labels to filter
    :return: a new image with only the largest connected component for each label
    """
    # Initialize the output image
    output_image = sitk.Image(image.GetSize(), image.GetPixelID())
    output_image.CopyInformation(image)

    for label in labels_of_interest:
        # Create a binary mask for the current label
        binary_mask = sitk.BinaryThreshold(
            image, lowerThreshold=label, upperThreshold=label
        )

        # Find connected components
        connected_components = sitk.ConnectedComponent(binary_mask)

        # Get label statistics
        label_stats = sitk.LabelShapeStatisticsImageFilter()
        label_stats.Execute(connected_components)

        # Find the label of the largest component
        largest_label = max(
            label_stats.GetLabels(), key=lambda i: label_stats.GetPhysicalSize(i)
        )

        # Create a mask of the largest component
        largest_component = sitk.BinaryThreshold(
            connected_components,
            lowerThreshold=largest_label,
            upperThreshold=largest_label,
        )

        # Add this component to the output image
        output_image = sitk.Add(output_image, sitk.Multiply(largest_component, label))

    return output_image
